import json
import logging
import time

from tqdm import tqdm

from gfypy.client.abstract_client import AbstractGfypy
from gfypy.const import GFYCAT_URL, FILEDROP_ENDPOINT
from gfypy.exceptions import GfypyAuthException, GfypyApiException
from gfypy.gfy import Gfy
from gfypy.http import SyncHttpClient
from gfypy.route import Route, CustomRoute

logger = logging.getLogger(__name__)


class Gfypy(AbstractGfypy):
    def __init__(self, client_id, client_secret, auth_file_path, **kwargs):
        super().__init__(client_id, client_secret, auth_file_path, **kwargs)
        self._http = SyncHttpClient(client_id, client_secret)

    def authenticate(self):
        if not self._auth_file_path.is_file():
            logger.info(
                'Credentials file "%s" does not exist. Creating it now.',
                self._auth_file_path,
            )
            with open(self._auth_file_path, "w") as auth_file:
                auth_file.write(json.dumps(self._http.creds))

        with open(self._auth_file_path, "r") as auth_file:
            self._http.creds = json.loads(auth_file.read())

            try:
                self._http.refresh_oauth_token(refresh=False)
            except GfypyAuthException as e:
                if e.code == "InvalidRefreshToken":
                    self._initial_auth()
                    self._auth_to_disk()
                else:
                    raise

    def close(self):
        self._http.close()

    def _initial_auth(self):
        self._http.get_oauth_token(self._get_oauth_code())

    def upload_from_file(
        self,
        filename,
        title="",
        tags=None,
        keep_audio=True,
        check_duplicate=False,
        check_upload=True,
    ):
        tags = tags or []

        key = self._get_key(title, tags, keep_audio, check_duplicate)
        payload = {"key": key}

        with open(filename, "rb") as file:
            files = {"file": (key, file.read())}

        self._http.request(
            CustomRoute("POST", FILEDROP_ENDPOINT),
            data=payload,
            files=files,
            no_auth=True,
        )

        if check_upload:
            status = self._check_upload_status(key)

            progress = tqdm(total=self.MAX_CHECKS)
            num_checks = 0
            while status["task"] != "complete":
                if num_checks > self.MAX_CHECKS:
                    # the gfycat was likely uploaded correctly, but gfycat is not sending 'task': 'complete'
                    break
                status = self._check_upload_status(key)
                progress.update(1)
                num_checks += 1
                time.sleep(3)

            progress.close()

            try:
                gfy = self.get_gfycat(key)

                logger.info(
                    "\n%s has been uploaded as %s/%s.", filename, GFYCAT_URL, key
                )
                return gfy
            except GfypyApiException:
                logger.info(
                    "\n%s has probably been uploaded as %s/%s, but the check was unsuccessful.",
                    filename,
                    GFYCAT_URL,
                    key,
                )
                return None
        else:
            logger.info(
                "\n%s has been uploaded as %s/%s; checks have been skipped.",
                filename,
                GFYCAT_URL,
                key,
            )

    def get_followers(self, fetch_userdata=False):
        resp = self._http.request(Route("GET", "/me/followers"))

        if not fetch_userdata:
            return resp["followers"]
        else:
            users = []
            for follower in resp["followers"]:
                user = self.get_user(follower["follower_id"])
                user["follow_date"] = follower["follow_date"]
                users.append(user)

            return users

    def get_user_feed(
        self, user_id=None, limit=100, sort_by=None, desc=True, filter_predicate=None
    ):
        if limit % 100 != 0 and limit >= 0:
            print("Limit needs to be divisible by 100. Rounding up.")

        gfycats = []
        i = 0
        cursor = ""

        if user_id is None:
            route = Route("GET", "/me/gfycats")
        else:
            route = Route("GET", "/users/{id}/gfycats", id=user_id)

        progress = tqdm(total=limit)

        while i < limit or limit < 0:
            resp = self._http.request(route, params={"count": 100, "cursor": cursor})

            cursor = resp["cursor"]
            new_gfys = Gfy.from_dict_list(self, resp["gfycats"])
            gfycats.extend(new_gfys)
            progress.update(len(new_gfys))

            if not cursor:
                print("Got no new entries from Gfycat. Stopping here.")
                break

            i = len(gfycats)

        progress.close()

        if filter_predicate:
            gfycats = [g for g in gfycats if filter_predicate(g)]

        if sort_by:
            gfycats = sorted(
                gfycats, key=lambda gfy: getattr(gfy, sort_by), reverse=desc
            )

        return gfycats



from .client import Gfypy

try:
    from .client import AsyncGfypy
except ImportError:
    pass

from .exceptions import GfypyApiException, GfypyAuthException, GfypyException
from .gfy import Gfy
from .helpers import is_pending
from .user import User
from .version import __version__

__all__ = [
    "AsyncGfypy",
    "Gfypy",
    "GfypyApiException",
    "GfypyAuthException",
    "GfypyException",
    "Gfy",
    "is_pending",
    "User",
    "__version__",
]



import json

import requests

from gfypy.const import REDIRECT_URI
from gfypy.exceptions import GfypyAuthException, GfypyApiException
from gfypy.http.abstract_http import AbstractHttpClient
from gfypy.route import Route


class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class SyncHttpClient(AbstractHttpClient):
    def __init__(self, client_id, client_secret):
        self._client_id = client_id
        self._client_secret = client_secret
        self._session = requests.Session()
        self._auth = None
        self.creds = {
            "access_token": None,
            "expires_in": None,
            "refresh_token": None,
            "refresh_token_expires_in": None,
            "resource_owner": None,
        }

    def close(self):
        self._session.close()

    def request(self, route, **kwargs):
        no_auth = kwargs.pop("no_auth", False)
        refresh = kwargs.pop("refresh", True)

        if self._auth is not None and not no_auth:
            resp = self._session.request(
                route.method, route.url, **kwargs, auth=self._auth
            )
        else:
            resp = self._session.request(route.method, route.url, **kwargs)

        content_type = resp.headers.get("content-type")

        if content_type == "application/json":
            content = resp.json()
        else:
            content = resp.text

        if 200 <= resp.status_code < 300:
            return content
        elif resp.status_code in [401, 403]:
            if refresh:
                # try to refresh the oauth token in case it's become invalid
                self.refresh_oauth_token()

                return self.request(route, **kwargs, refresh=False)
            else:
                if "message" in content:
                    raise GfypyAuthException(content["message"], resp.status_code, None)
                else:
                    raise GfypyAuthException(
                        content["errorMessage"]["description"],
                        resp.status_code,
                        content["errorMessage"]["code"],
                    )
        else:
            raise GfypyApiException(
                content["errorMessage"] if "message" in content else content,
                resp.status_code,
            )

    def get_oauth_token(self, code):
        payload = {
            "code": code,
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "authorization_code",
            "redirect_uri": REDIRECT_URI,
        }

        resp = self.request(
            Route("POST", "/oauth/token"),
            data=json.dumps(payload),
            headers={"content-type": "application/json"},
        )

        self.creds = resp
        self.auth = resp["access_token"]

    def refresh_oauth_token(self, **kwargs):
        payload = {
            "refresh_token": self.creds["refresh_token"],
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "refresh",
        }

        resp = self.request(
            Route("POST", "/oauth/token"),
            data=json.dumps(payload),
            headers={"content-type": "application/json"},
            **kwargs,
        )

        self.creds = resp
        self.auth = resp["access_token"]

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, token):
        self._auth = BearerAuth(token)



import json

import aiohttp

from gfypy.const import REDIRECT_URI
from gfypy.exceptions import GfypyAuthException, GfypyApiException
from gfypy.http.abstract_http import AbstractHttpClient
from gfypy.route import Route


class BearerAuth(aiohttp.BasicAuth):
    def __init__(self, token):
        self.token = token

    def encode(self):
        return "Bearer " + self.token


class AsyncHttpClient(AbstractHttpClient):
    def __init__(self, client_id, client_secret):
        self._client_id = client_id
        self._client_secret = client_secret
        self._session = aiohttp.ClientSession()
        self._auth = None
        self.creds = {
            "access_token": None,
            "expires_in": None,
            "refresh_token": None,
            "refresh_token_expires_in": None,
            "resource_owner": None,
        }

    async def close(self):
        await self._session.close()

    async def request(self, route, **kwargs):
        no_auth = kwargs.pop("no_auth", False)
        refresh = kwargs.pop("refresh", True)

        async with self._session.request(
            route.method, route.url, **kwargs, auth=self._auth if not no_auth else None
        ) as resp:
            content_type = resp.headers.get("content-type")
            if content_type == "application/json":
                content = await resp.json()
            else:
                content = await resp.text()

            if 200 <= resp.status < 300:
                return content
            elif resp.status in [401, 403]:
                if refresh:
                    # try to refresh the oauth token in case it's become invalid
                    await self.refresh_oauth_token()

                    return await self.request(route, **kwargs, refresh=False)
                else:
                    if "message" in content:
                        raise GfypyAuthException(content["message"], resp.status, None)
                    else:
                        raise GfypyAuthException(
                            content["errorMessage"]["description"],
                            resp.status,
                            content["errorMessage"]["code"],
                        )
            else:
                raise GfypyApiException(
                    content["errorMessage"] if "message" in content else content,
                    resp.status,
                )

    async def get_oauth_token(self, code):
        payload = {
            "code": code,
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "authorization_code",
            "redirect_uri": REDIRECT_URI,
        }

        resp = await self.request(
            Route("POST", "/oauth/token"),
            data=json.dumps(payload),
            headers={"content-type": "application/json"},
        )

        self.creds = resp
        self.auth = resp["access_token"]

    async def refresh_oauth_token(self, **kwargs):
        payload = {
            "refresh_token": self.creds["refresh_token"],
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "refresh",
        }

        resp = await self.request(
            Route("POST", "/oauth/token"),
            data=json.dumps(payload),
            headers={"content-type": "application/json"},
            **kwargs,
        )

        self.creds = resp
        self.auth = resp["access_token"]

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, token):
        self._auth = BearerAuth(token)



import asyncio
import json
import logging

from aiohttp import FormData

from gfypy.client.abstract_client import AbstractGfypy
from gfypy.const import FILEDROP_ENDPOINT, GFYCAT_URL
from gfypy.exceptions import GfypyAuthException, GfypyApiException
from gfypy.gfy import Gfy
from gfypy.http import AsyncHttpClient
from gfypy.route import CustomRoute
from gfypy.route import Route

logger = logging.getLogger(__name__)


class AsyncGfypy(AbstractGfypy):
    def __init__(self, client_id, client_secret, auth_file_path, **kwargs):
        super().__init__(client_id, client_secret, auth_file_path, **kwargs)
        self._http = AsyncHttpClient(client_id, client_secret)

    async def authenticate(self):
        if not self._auth_file_path.is_file():
            logging.info(
                'Credentials file "%s" does not exist. Creating it now.',
                self._auth_file_path,
            )
            with open(self._auth_file_path, "w") as auth_file:
                auth_file.write(json.dumps(self._http.creds))

        with open(self._auth_file_path, "r") as auth_file:
            self._http.creds = json.loads(auth_file.read())

            try:
                await self._http.refresh_oauth_token(refresh=False)
            except GfypyAuthException as e:
                if e.code == "InvalidRefreshToken":
                    await self._initial_auth()
                    self._auth_to_disk()
                else:
                    raise

    async def close(self):
        await self._http.close()

    async def _initial_auth(self):
        await self._http.get_oauth_token(self._get_oauth_code())

    async def upload_from_file(
        self,
        filename,
        title="",
        tags=None,
        keep_audio=True,
        check_duplicate=False,
        check_upload=True,
    ):
        tags = tags or []

        key = await self._get_key(title, tags, keep_audio, check_duplicate)

        data = FormData()
        data.add_field("key", key)

        with open(filename, "rb") as file:
            data.add_field("file", file, filename=filename)

            await self._http.request(
                CustomRoute("POST", FILEDROP_ENDPOINT), data=data, no_auth=True
            )

        if check_upload:
            status = await self._check_upload_status(key)

            num_checks = 0
            while status["task"] != "complete":
                if num_checks > self.MAX_CHECKS:
                    # the gfycat was likely uploaded correctly, but gfycat is not sending 'task': 'complete'
                    break
                status = await self._check_upload_status(key)
                num_checks += 1
                await asyncio.sleep(3)

            try:
                gfy = await self.get_gfycat(key)

                logger.info(
                    "\n%s has been uploaded as %s/%s.", filename, GFYCAT_URL, key
                )
                return gfy
            except GfypyApiException:
                logger.info(
                    "\n%s has probably been uploaded as %s/%s, but the check was unsuccessful.",
                    filename,
                    GFYCAT_URL,
                    key,
                )
                return None
        else:
            logger.info(
                "\n%s has been uploaded as %s/%s; checks have been skipped.",
                filename,
                GFYCAT_URL,
                key,
            )

    async def user_feed_generator(self, user_id=None, per_request=100):
        if not 20 <= per_request <= 100:
            raise ValueError("Number per request needs to be between 20 and 100.")

        cursor = ""

        if user_id is None:
            route = Route("GET", "/me/gfycats")
        else:
            route = Route("GET", "/users/{id}/gfycats", id=user_id)

        while True:
            resp = await self._http.request(
                route, params={"count": per_request, "cursor": cursor}
            )

            cursor = resp["cursor"]
            new_gfys = Gfy.from_dict_list(self, resp["gfycats"])
            for new_gfy in new_gfys:
                yield new_gfy

            if not cursor:
                logger.info("Got no new entries from Gfycat. Stopping here.")
                break

    async def get_user_feed(
        self, user_id=None, limit=100, sort_by=None, desc=True, filter_predicate=None
    ):
        if limit % 100 != 0 and limit >= 0:
            logger.warning("Limit needs to be divisible by 100. Rounding up.")

        gfycats = []

        async for gfy in self.user_feed_generator(user_id=user_id, per_request=100):
            gfycats.append(gfy)

            if len(gfycats) >= limit >= 0:
                break

        if filter_predicate:
            gfycats = [g for g in gfycats if filter_predicate(g)]

        if sort_by:
            gfycats = sorted(
                gfycats, key=lambda gfy: getattr(gfy, sort_by), reverse=desc
            )

        return gfycats



import asyncio
import inspect
import json
import time
import webbrowser
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from threading import Thread
from urllib.parse import urlparse, urlencode

from gfypy.const import AUTH_ENDPOINT, REDIRECT_URI
from gfypy.exceptions import GfypyException
from gfypy.gfy import Gfy
from gfypy.route import Route
from gfypy.user import User


class Promise:
    def __init__(self, coro):
        self.coro = coro
        self._is_coro = inspect.iscoroutine(coro)

    def __await__(self):
        if self._is_coro:
            return self.coro.__await__()
        return asyncio.sleep(0, result=self.coro).__await__()

    def then(self, later):
        if self._is_coro:

            async def chain():
                value = await self.coro
                return await Promise(later(value))

            return chain()
        else:
            return later(self.coro)


class AuthCallbackRequestHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(bytes("<h2>You may close this window now!</h2>", "UTF-8"))
        query = urlparse(self.path).query
        query_components = dict(qc.split("=") for qc in query.split("&"))
        code = query_components["code"]
        self.server.code = code


class AbstractGfypy:
    MAX_TAGS = 20
    MAX_CHECKS = 30

    def __init__(self, client_id, client_secret, auth_file_path, headless=False):
        self._client_id = client_id
        self._client_secret = client_secret
        self._auth_file_path = Path(auth_file_path)
        self._headless = headless

    def _get_oauth_code(self):
        """
        Gets authorization token
        """
        params = {
            "client_id": self._client_id,
            "scope": "all",
            "state": "Gfypy",
            "response_type": "code",
            "redirect_uri": REDIRECT_URI,
        }

        auth_url = f"{AUTH_ENDPOINT}?{urlencode(params)}"

        if self._headless:
            return input(
                f"Please open the following URL in a web browser, authorize the application and enter the "
                f"code from the code param (?code=...).\n{auth_url}\nCode: "
            )
        else:
            server = ThreadingHTTPServer(
                ("localhost", 8000), AuthCallbackRequestHandler
            )
            server.code = None
            server_thread = Thread(target=server.serve_forever)
            server_thread.daemon = False
            server_thread.start()

            webbrowser.open(auth_url)

            while server.code is None:
                time.sleep(1)

            assassin = Thread(target=server.shutdown)
            assassin.daemon = True
            assassin.start()

            return server.code

    def _auth_to_disk(self):
        with open(self._auth_file_path, "w") as auth_file:
            auth_file.write(json.dumps(self._http.creds))

    def _check_upload_status(self, gfy_key):
        return self._http.request(
            Route("GET", "/gfycats/fetch/status/{gfy_key}", gfy_key=gfy_key)
        )

    def get_me(self):
        return Promise(self._http.request(Route("GET", "/me"))).then(
            lambda r: User.from_dict(self._http, r)
        )

    def _get_key(self, title="", tags=None, keep_audio=True, check_duplicate=False):
        tags = tags or []

        if len(tags) > self.MAX_TAGS:
            raise GfypyException(
                f"Too many tags. Supplied {len(tags)}, max. {self.MAX_TAGS}."
            )

        payload = {
            "title": title,
            "tags": tags,
            "keepAudio": keep_audio,
            "noMd5": not check_duplicate,
        }

        return Promise(
            self._http.request(
                Route("POST", "/gfycats"),
                data=json.dumps(payload),
                headers={"content-type": "application/json"},
            )
        ).then(lambda r: r["gfyname"])

    def get_user_feed(self, user_id, **kwargs):
        raise NotImplementedError

    def get_own_feed(self, **kwargs):
        return self.get_user_feed(**kwargs)

    def get_gfycat(self, _id):
        return Promise(self._http.request(Route("GET", "/gfycats/{id}", id=_id))).then(
            lambda r: Gfy.from_dict(self._http, r["gfyItem"])
        )

    def get_user(self, _id):
        return Promise(self._http.request(Route("GET", "/users/{id}", id=_id))).then(
            lambda r: User.from_dict(self._http, json.loads(r))
        )



from urllib.parse import quote


class Route:
    BASE = "https://api.gfycat.com/v1"

    def __init__(self, method, path, **parameters):
        self.method = method
        self.path = path
        url = self.BASE + self.path
        if parameters:
            self.url = url.format(
                **{
                    k: quote(v) if isinstance(v, str) else v
                    for k, v in parameters.items()
                }
            )
        else:
            self.url = url


class CustomRoute:
    def __init__(self, method, base, path=None):
        self.method = method
        self.base = base
        self.path = path
        if path is not None:
            self.url = self.base + self.path
        else:
            self.url = self.base



import json
from datetime import datetime

from .route import Route

from dataclasses import dataclass
import typing


@dataclass
class ContentUrl:
    def __init__(self, **kwargs):
        self.url: str = kwargs.pop("url", None)
        self.size: int = kwargs.pop("size", None)
        self.height: int = kwargs.pop("height", None)
        self.width: int = kwargs.pop("width", None)


@dataclass
class UserData:
    def __init__(self, **kwargs):
        self.name: str = kwargs.pop("name", None)
        self.profile_image_url: str = kwargs.pop("profileImageUrl", None)
        self.url: str = kwargs.pop("url", None)
        self.username: str = kwargs.pop("username", None)
        self.followers: int = kwargs.pop("followers", None)
        self.subscription: int = kwargs.pop("subscription", None)
        self.following: int = kwargs.pop("following", None)
        self.profile_url: str = kwargs.pop("profileUrl", None)
        self.views: int = kwargs.pop("views", None)
        self.verified: bool = kwargs.pop("verified", None)


@dataclass
class ContentUrls:
    def __init__(self, **kwargs):
        self.max2_mb_gif = (
            ContentUrl(**kwargs.pop("max2mbGif")) if "max2mbGif" in kwargs else None
        )
        self.webp = ContentUrl(**kwargs.pop("webp")) if "webp" in kwargs else None
        self.max1_mb_gif = (
            ContentUrl(**kwargs.pop("max1mbGif")) if "max1mbGif" in kwargs else None
        )
        self._100_px_gif = (
            ContentUrl(**kwargs.pop("100pxGif")) if "100pxGif" in kwargs else None
        )
        self.mobile_poster = (
            ContentUrl(**kwargs.pop("mobilePoster"))
            if "mobilePoster" in kwargs
            else None
        )
        self.mp4 = ContentUrl(**kwargs.pop("mp4")) if "mp4" in kwargs else None
        self.webm = ContentUrl(**kwargs.pop("webm")) if "webm" in kwargs else None
        self.max5_mb_gif = (
            ContentUrl(**kwargs.pop("max5mbGif")) if "max5mbGif" in kwargs else None
        )
        self.large_gif = (
            ContentUrl(**kwargs.pop("largeGif")) if "largeGif" in kwargs else None
        )
        self.mobile = ContentUrl(**kwargs.pop("mobile")) if "mobile" in kwargs else None


@dataclass
class Gfy(dict):
    def __init__(self, http, **kwargs):
        super().__init__(kwargs)
        self._http = http
        self._source = kwargs

        self.content_urls = ContentUrls(**kwargs.pop("content_urls", {}))
        self.user_data = UserData(**kwargs.pop("userData", {}))

        # explicit cast to int required because the API sometimes returns a str
        self.likes: int = int(kwargs.pop("likes", 0))
        self.dislikes: int = int(kwargs.pop("dislikes", 0))
        self.gfy_number: int = int(kwargs.pop("gfyNumber", 0))

        self.title = kwargs.pop("title", None)
        self.tags: typing.Set[str] = set(kwargs.pop("tags", None))
        self.language_categories: typing.List[typing.Any] = kwargs.pop(
            "languageCategories", []
        )
        self.domain_whitelist: typing.List[typing.Any] = kwargs.pop(
            "domainWhitelist", []
        )
        self.geo_whitelist: typing.List[typing.Any] = kwargs.pop("geoWhitelist", [])
        self.published: int = kwargs.pop("published", None)
        self.nsfw: int = kwargs.pop("nsfw", None)
        self.gatekeeper: int = kwargs.pop("gatekeeper", None)
        self.mp4_url: str = kwargs.pop("mp4Url", None)
        self.gif_url: str = kwargs.pop("gifUrl", None)
        self.webm_url: str = kwargs.pop("webmUrl", None)
        self.webp_url: str = kwargs.pop("webpUrl", None)
        self.mobile_url: str = kwargs.pop("mobileUrl", None)
        self.mobile_poster_url: str = kwargs.pop("mobilePosterUrl", None)
        self.extra_lemmas: str = kwargs.pop("extraLemmas", None)
        self.thumb100_poster_url: str = kwargs.pop("thumb100PosterUrl", None)
        self.mini_url: str = kwargs.pop("miniUrl", None)
        self.gif100_px: str = kwargs.pop("gif100px", None)
        self.mini_poster_url: str = kwargs.pop("miniPosterUrl", None)
        self.max5_mb_gif: str = kwargs.pop("max5mbGif", None)
        self.max2_mb_gif: str = kwargs.pop("max2mbGif", None)
        self.max1_mb_gif: str = kwargs.pop("max1mbGif", None)
        self.poster_url: str = kwargs.pop("posterUrl", None)
        self.language_text: str = kwargs.pop("languageText", None)
        self.views: int = kwargs.pop("views", None)
        self.user_name: str = kwargs.pop("userName", None)
        self.description: str = kwargs.pop("description", None)
        self.sitename: str = kwargs.pop("sitename", None)
        self.has_transparency: bool = kwargs.pop("hasTransparency", None)
        self.has_audio: bool = kwargs.pop("hasAudio", None)
        self.gfy_id: str = kwargs.pop("gfyId", None)
        self.gfy_name: str = kwargs.pop("gfyName", None)
        self.width: int = kwargs.pop("width", None)
        self.height: int = kwargs.pop("height", None)
        self.frame_rate: float = kwargs.pop("frameRate", None)
        self.num_frames: int = kwargs.pop("numFrames", None)
        self.mp4_size: int = kwargs.pop("mp4Size", None)
        self.webm_size: int = kwargs.pop("webmSize", None)
        self.create_date: datetime = datetime.fromtimestamp(
            kwargs.pop("createDate", None)
        )
        self.source: int = kwargs.pop("source", None)
        self.gfy_slug: str = kwargs.pop("gfySlug", None)
        self.md5: str = kwargs.pop("md5", None)
        self.rating: typing.Any = kwargs.pop("rating", None)
        self.avg_color: str = kwargs.pop("avgColor", None)
        self.user_display_name: str = kwargs.pop("userDisplayName", None)
        self.user_profile_image_url: str = kwargs.pop("userProfileImageUrl", None)

    @staticmethod
    def from_dict(http, source):
        return Gfy(http, **source)

    @staticmethod
    def from_dict_list(http, source):
        return [Gfy.from_dict(http, gfy) for gfy in source]

    def set_title(self, new_title):
        payload = {"value": new_title}

        return self._http.request(
            Route("PUT", "/me/gfycats/{id}/title", id=self["gfyId"]),
            data=json.dumps(payload),
        )

    def delete_title(self):
        return self._http.request(
            Route("DELETE", "/me/gfycats/{id}/title", id=self["gfyId"])
        )

    def delete(self):
        return self._http.request(Route("DELETE", "/me/gfycats/{id}", id=self["gfyId"]))



import typing
from dataclasses import dataclass
from datetime import datetime


@dataclass
class User(dict):
    def __init__(self, http, **kwargs):
        super().__init__(kwargs)
        self._http = http
        self._source = kwargs

        # explicit cast to int required because the API sometimes returns a str
        self.following: int = (
            int(kwargs.pop("following")) if "following" in kwargs else None
        )
        self.followers: int = (
            int(kwargs.pop("followers")) if "followers" in kwargs else None
        )
        self.views: int = int(kwargs.pop("views")) if "views" in kwargs else None
        self.create_date: datetime = (
            datetime.fromtimestamp(int(kwargs.pop("createDate")))
            if "createDate" in kwargs
            else None
        )
        self.published_gfycats: int = (
            int(kwargs.pop("publishedGfycats"))
            if "publishedGfycats" in kwargs
            else None
        )
        self.total_gfycats: int = (
            int(kwargs.pop("totalGfycats")) if "totalGfycats" in kwargs else None
        )
        self.total_bookmarks: int = (
            int(kwargs.pop("totalBookmarks")) if "totalBookmarks" in kwargs else None
        )
        self.total_albums: int = (
            int(kwargs.pop("totalAlbums")) if "totalAlbums" in kwargs else None
        )
        self.published_albums: int = (
            int(kwargs.pop("publishedAlbums")) if "publishedAlbums" in kwargs else None
        )

        self.subscription: int = kwargs.pop("subscription", None)
        self.verified: bool = kwargs.pop("verified", None)
        self.iframe_profile_image_visible: bool = kwargs.pop(
            "iframeProfileImageVisible", None
        )
        self.url: str = kwargs.pop("url", None)
        self.userid: str = kwargs.pop("userid", None)
        self.username: str = kwargs.pop("username", None)
        self.description: str = kwargs.pop("description", None)
        self.name: str = kwargs.pop("name", None)
        self.profile_image_url: str = kwargs.pop("profileImageUrl", None)
        self.email: str = kwargs.pop("email", None)
        self.canonical_username: str = kwargs.pop("canonicalUsername", None)
        self.email_verified: bool = kwargs.pop("emailVerified", None)
        self.viewing_preference: int = kwargs.pop("viewingPreference", None)
        self.domain_whitelist: typing.List[typing.Any] = kwargs.pop(
            "domainWhitelist", []
        )
        self.geo_whitelist: typing.List[typing.Any] = kwargs.pop("geoWhitelist", [])

    @staticmethod
    def from_dict(http, source):
        return User(http, **source)

    @staticmethod
    def from_dict_list(http, source):
        return [User.from_dict(http, user) for user in source]



from setuptools import setup, find_packages

exec(open(r"gfypy/version.py").read())

dependencies = ["requests", "tqdm"]

setup(
    name="gfypy",
    version=__version__,  # noqa
    description="Python wrapper for the Gfycat API",
    long_description="Provides an easy to use interface to the API at "
    "https://developers.gfycat.com/api/.",
    url="https://github.com/noahkw/gfypy",
    author="Noah K W",
    author_email="noah@dreckbu.de",
    license="MIT",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=["gfycat api wrapper"],
    packages=find_packages(),
    install_requires=dependencies,
    setup_requires=dependencies,
    extras_require={"async": ["aiohttp"]},
    scripts=["bin/gfy-uploader"],
)



__all__ = ("GfypyException", "GfypyApiException", "GfypyAuthException")


class GfypyException(Exception):
    def __init__(self, error_msg):
        self.error_msg = error_msg


class GfypyApiException(GfypyException):
    def __init__(self, error_msg, status_code):
        super().__init__(error_msg)
        self.status_code = status_code

    def __str__(self):
        return f"The Gfycat API responded with a {self.status_code}: {self.error_msg}"


class GfypyAuthException(GfypyApiException):
    def __init__(self, error_msg, status_code, code):
        super().__init__(error_msg, status_code)
        self.code = code

    def __str__(self):
        return f"The Gfycat API responded with a {self.status_code} / {self.code}: {self.error_msg}"
